#ifndef FACTORIZATIONCONTROLLER_H
#define FACTORIZATIONCONTROLLER_H

#include <QObject>
#include <QTimer>
#include <qqmlintegration.h>
#include <qtmetamacros.h>

/**
 * @class FactorizationController
 * @brief A class that performs integer factorization in a step-by-step manner.
 *
 * This class uses a timer to progressively factorize a given number, allowing
 * for real-time updates on progress and discovered factors. It is designed
 * to be used in a Qt-based QML application.
 */
class FactorizationController : public QObject {
    Q_OBJECT
    QML_ELEMENT
    Q_PROPERTY(bool isRunning READ isRunning NOTIFY isRunningChanged)
    Q_PROPERTY(bool isFinished READ isFinished NOTIFY isFinishedChanged)
    Q_PROPERTY(bool isPaused READ isPaused NOTIFY isPausedChanged)
    Q_PROPERTY(long long number READ number NOTIFY numberChanged)
    Q_PROPERTY(long long currentFactor READ currentFactor NOTIFY currentFactorChanged)
    Q_PROPERTY(int progress READ progress NOTIFY progressChanged)
    Q_PROPERTY(QList<long long> factors READ factors NOTIFY factorsChanged)

  public:
    explicit FactorizationController(QObject* parent = nullptr);

    // region: Getters

    /**
     * @brief Gets the current progress of factorization.
     * @return The progress percentage (0-100).
     */
    int progress() const;

    /**
     * @brief Checks if the factorization process is running.
     * @return True if the factorization is in progress, otherwise false.
     *
     * This just informs whether the task is active (scheduled to run), not whether it's
     * running at this very moment, as this implementation doesn't use threads.
     */
    bool isRunning() const;

    /**
     * @brief Checks if the factorization process is done.
     * @return True if the factorization is completed, otherwise false.
     *
     * This will only be true if the factorization process finished without being
     * explicitly stopped ahead of time (e.g. when the computation is done and we
     * have all the factors).
     */
    bool isFinished() const;

    /**
     * @brief Checks if the factorization process is currently paused.
     * @return True if the factorization process is paused, otherwise false.
     *
     * This function differs from just !isRunning, as this will only be set when
     * the computation is currently paused, but can still be resumed. E.g. the
     * computation was already started at some point, and stopped (paused) later,
     * without having been reset.
     *
     * This being true is a prerequisite to resuming.
     */
    bool isPaused() const;

    /**
     * @brief Gets the original number being factorized.
     * @return The number being factorized.
     */
    long long number() const;

    /**
     * @brief Gets the current factor being tested.
     * @return The current factor.
     *
     * The factorization process is implemented in a way that simply checks whether a number
     * is a factor, going one by one. This returns the current factor (to be checked in the
     * next iteration).
     */
    long long currentFactor() const;

    /**
     * @brief Gets the list of discovered factors (may not be complete).
     * @return A QList of extracted prime factors.
     *
     * This returns the factors that were already discovered Note that this list might not
     * be complete (if the computation is still running). If you need to know whether you
     * can trust this result, check isFinished first.
     */
    QList<long long> factors() const;

    // endregion
    // region: Functions

    /**
     * @brief Starts the factorization process for a given number.
     * @param number The number to be factorized.
     *
     * This will start a 0-tick timer, which triggers a partial computation on each run,
     * to let the event loop cycle in between. This implementation does NOT rely on threads.
     */
    Q_INVOKABLE void start(long long number);

    /**
     * @brief Stops the ongoing factorization process.
     *
     * This will stop the timer, meaning the computation will no longer get triggered
     * by the event loop, essentially pausing it. Note that this does not reset the already
     * computed details, which means the computation can be resumed later, allowing for
     * inspection of the currently computed values.
     */
    Q_INVOKABLE void stop();

    /**
     * @brief Resumes a previously stopped factorization process.
     */
    Q_INVOKABLE void resume();

    /**
     * @brief Resets the factorization process data.
     *
     * This will clear the current progress (if any). If the factorization process is running,
     * this will stop it. If the factorization process is paused, this will reset isPaused,
     * forcing a restart.
     */
    Q_INVOKABLE void reset();

    // endregion

  signals:
    void isRunningChanged();
    void isFinishedChanged();
    void numberChanged();
    void currentFactorChanged();
    void factorsChanged();
    void progressChanged();
    void isPausedChanged();

  private slots:
    void onTimerTick();

  private:
    bool m_isRunning; ///< Indicates whether factorization is active.
    bool m_isFinished; ///< Indicates whether the factorization process is done.
    bool m_isPaused; ///< Indicates whether the factorization process is paused (can be resumed).
    long long m_currentFactNumber; ///< The number currently being factorized.
    long long m_originalNumber; ///< The original input number.
    long long m_currentFactor; ///< The current divisor being checked.
    long long m_stopFactor; ///< The stopping limit (square root of original number, saved for efficiency).
    QList<long long> m_factors; ///< List of discovered prime factors.
    QTimer m_timer; ///< Timer for stepwise execution.

    /**
     * @brief Performs one step of the factorization process.
     */
    void factorize();
};

#endif // FACTORIZATIONCONTROLLER_H
